from pymongo import ReadPreference
import pytest
from mongomock_motor import AsyncMongoMockClient


@pytest.mark.anyio
async def test_attrs():
    client = AsyncMongoMockClient()
    database = client.get_database('tests')
    collection = database.get_collection('test')
    await collection.insert_one({'a': 1})

    assert await client['tests']['test'].find_one(projection={'_id': 0, 'a': 1}) == {'a': 1}
    assert await client.tests.test.find_one(projection={'_id': 0, 'a': 1}) == {'a': 1}
    assert await collection.find_one(projection={'_id': 0, 'a': 1}) == {'a': 1}

    database = client.get_database('tests', read_preference=ReadPreference.SECONDARY)
    assert await database.test.find_one(projection={'_id': 0, 'a': 1}) == {'a': 1}



import pytest
from mongomock_motor import AsyncMongoMockClient


@pytest.mark.anyio
async def test_lists():
    client = AsyncMongoMockClient()

    collection1 = client['tests-1']['test-1']
    await collection1.insert_one({'a': 1})

    collection2 = client['tests-2']['test-2']
    await collection2.insert_one({'a': 2})

    collection2 = client['tests-2']['test-2-copy']
    await collection2.insert_one({'a': 2})

    assert await client.list_database_names() == ['tests-1', 'tests-2']
    assert await client['tests-1'].list_collection_names() == ['test-1']
    assert await client['tests-2'].list_collection_names() == ['test-2', 'test-2-copy']



from mongomock import DuplicateKeyError, helpers


try:
    from beanie.odm.fields import ExpressionField
except ModuleNotFoundError:
    ExpressionField = None


def _provide_error_details(collection, data, exception):
    if not isinstance(exception, DuplicateKeyError):
        return exception

    for index in collection._store.indexes.values():
        if not index.get('unique'):
            continue

        is_sparse = index.get('sparse')

        find_kwargs = {}
        for key, _ in index.get('key'):
            try:
                find_kwargs[key] = helpers.get_value_by_dot(data, key)
            except KeyError:
                find_kwargs[key] = None

        if is_sparse and set(find_kwargs.values()) == {None}:
            continue

        found_documents = list(collection._iter_documents(find_kwargs))
        if len(found_documents) > 0:
            return DuplicateKeyError(
                'E11000 Duplicate Key Error',
                11000,
                {
                    'keyValue': find_kwargs,
                    'keyPattern': dict(index.get('key')),
                },
                None,
            )

    return exception


def _patch_insert(collection):
    """
    Adds details with 'keyPattern' and 'keyValue' when
    raising DuplicateKeyError from _insert
    https://github.com/mongomock/mongomock/issues/773
    """
    _insert = collection._insert

    def insert(data, *args, **kwargs):
        try:
            return _insert(data, *args, **kwargs)
        except DuplicateKeyError as exc:
            raise _provide_error_details(collection, data, exc)

    collection._insert = insert

    return collection


def _normalize_strings(obj):
    if isinstance(obj, list):
        return [_normalize_strings(v) for v in obj]

    if isinstance(obj, dict):
        return {_normalize_strings(k): _normalize_strings(v) for k, v in obj.items()}

    # make sure we won't fail while working with beanie
    if ExpressionField and isinstance(obj, ExpressionField):
        return str(obj)

    return obj


def _patch_iter_documents(collection):
    """
    When using beanie, keys can have "ExpressionField" type,
    that is inherited from "str". Looks like pymongo works ok
    with that, so we should be too.
    """
    _iter_documents = collection._iter_documents

    def iter_documents(filter):
        return _iter_documents(_normalize_strings(filter))

    collection._iter_documents = iter_documents

    return collection


def _patch_collection_internals(collection):
    collection = _patch_insert(collection)
    collection = _patch_iter_documents(collection)
    return collection


__all__ = ['_patch_collection_internals']



from functools import wraps
import importlib
from mongomock import MongoClient
from .patches import _patch_collection_internals


def masquerade_class(name):
    module_name, target_name = name.rsplit('.', 1)

    try:
        target = getattr(importlib.import_module(module_name), target_name)
    except Exception:
        return lambda cls: cls

    def decorator(cls):
        @wraps(target, updated=())
        class Wrapper(cls):
            @property
            def __class__(self):
                return target
        return Wrapper
    return decorator


def with_async_methods(source, async_methods):
    def decorator(cls):
        for method_name in async_methods:
            def make_wrapper(method_name):
                async def wrapper(self, *args, **kwargs):
                    proxy_source = self.__dict__.get(f'_{cls.__name__}{source}')
                    return getattr(proxy_source, method_name)(*args, **kwargs)
                return wrapper

            setattr(cls, method_name, make_wrapper(method_name))

        return cls

    return decorator


def with_cursor_chaining_methods(source, chaining_methods):
    def decorator(cls):
        for method_name in chaining_methods:
            def make_wrapper(method_name):
                def wrapper(self, *args, **kwargs):
                    proxy_source = self.__dict__.get(f'_{cls.__name__}{source}')
                    getattr(proxy_source, method_name)(*args, **kwargs)
                    return self
                return wrapper

            setattr(cls, method_name, make_wrapper(method_name))

        return cls

    return decorator


@masquerade_class('motor.motor_asyncio.AsyncIOMotorCursor')
@with_cursor_chaining_methods('__cursor', [
    'add_option',
    'allow_disk_use',
    'collation',
    'comment',
    'hint',
    'limit',
    'max_await_time_ms',
    'max_scan',
    'max_time_ms',
    'max',
    'min',
    'remove_option',
    'skip',
    'sort',
    'where',
])
class AsyncCursor():
    def __init__(self, cursor):
        self.__cursor = cursor

    def __getattr__(self, name):
        return getattr(self.__cursor, name)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.__cursor)
        except StopIteration:
            raise StopAsyncIteration()

    def clone(self):
        return AsyncCursor(self.__cursor.clone())

    async def distinct(self, *args, **kwargs):
        return self.__cursor.distinct(*args, **kwargs)

    async def to_list(self, *args, **kwargs):
        return list(self.__cursor)


@masquerade_class('motor.motor_asyncio.AsyncIOMotorLatentCommandCursor')
class AsyncLatentCommandCursor():
    def __init__(self, cursor):
        self.__cursor = cursor

    def __getattr__(self, name):
        return getattr(self.__cursor, name)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.__cursor)
        except StopIteration:
            raise StopAsyncIteration()

    async def to_list(self, *args, **kwargs):
        return list(self.__cursor)


@masquerade_class('motor.motor_asyncio.AsyncIOMotorCollection')
@with_async_methods('__collection', [
    'bulk_write',
    'count_documents',
    'count',  # deprecated
    'create_index',
    'create_indexes',
    'delete_many',
    'delete_one',
    'drop_index',
    'drop_indexes',
    'drop',
    'distinct',
    'ensure_index',
    'estimated_document_count',
    'find_and_modify',  # deprecated
    'find_one_and_delete',
    'find_one_and_replace',
    'find_one_and_update',
    'find_one',
    'index_information',
    'inline_map_reduce',
    'insert_many',
    'insert_one',
    'map_reduce',
    'options',
    'reindex',
    'rename',
    'replace_one',
    'save',
    'update_many',
    'update_one',
])
class AsyncMongoMockCollection():
    def __init__(self, database, collection):
        self.database = database
        self.__collection = collection

    def __eq__(self, other):
        return self.__collection == other.__collection

    def __getattr__(self, name):
        return getattr(self.__collection, name)

    def find(self, *args, **kwargs) -> AsyncCursor:
        return AsyncCursor(self.__collection.find(*args, **kwargs))

    def aggregate(self, *args, **kwargs) -> AsyncLatentCommandCursor:
        return AsyncLatentCommandCursor(self.__collection.aggregate(*args, **kwargs))


@masquerade_class('motor.motor_asyncio.AsyncIOMotorDatabase')
@with_async_methods('__database', [
    'create_collection',
    'dereference',
    'drop_collection',
    'list_collection_names',
    'validate_collection',
])
class AsyncMongoMockDatabase():
    def __init__(self, client, database, mock_build_info=None):
        self.client = client
        self.__database = database
        self.__build_info = mock_build_info or {'ok': 1.0, 'version': '5.0.5'}

    def get_collection(self, *args, **kwargs):
        return AsyncMongoMockCollection(
            self,
            _patch_collection_internals(
                self.__database.get_collection(*args, **kwargs),
            ),
        )

    def aggregate(self, *args, **kwargs) -> AsyncCursor:
        return AsyncCursor(self.__database.aggregate(*args, **kwargs))

    async def command(self, *args, **kwargs):
        try:
            return getattr(self.__database, 'command')(*args, **kwargs)
        except NotImplementedError:
            if args == ({'buildInfo': 1},) and not kwargs:
                return self.__build_info
            raise

    def __eq__(self, other):
        return self.__database == other.__database

    def __getitem__(self, name):
        return self.get_collection(name)

    def __getattr__(self, name):
        if name in dir(self.__database):
            return getattr(self.__database, name)

        return self.get_collection(name)


@masquerade_class('motor.motor_asyncio.AsyncIOMotorClient')
@with_async_methods('__client', [
    'drop_database',
    'list_database_names',
    'list_databases',
    'server_info',
])
class AsyncMongoMockClient():
    def __init__(self, *args, mock_build_info=None, **kwargs):
        self.__client = MongoClient(*args, **kwargs)
        self.__build_info = mock_build_info

    def get_database(self, *args, **kwargs):
        return AsyncMongoMockDatabase(
            self,
            self.__client.get_database(*args, **kwargs),
            mock_build_info=self.__build_info,
        )

    def __eq__(self, other):
        return self.__client == other.__client

    def __getitem__(self, name):
        return self.get_database(name)

    def __getattr__(self, name):
        if name in dir(self.__client):
            return getattr(self.__client, name)

        return self.get_database(name)



from enum import Enum

import pytest
from mongo_thingy import AsyncThingy, connect
from pymongo.errors import DuplicateKeyError

from mongomock_motor import AsyncMongoMockClient


@pytest.mark.anyio
async def test_mongo_thingy():
    connect(client_cls=AsyncMongoMockClient)

    class Something(AsyncThingy):
        pass

    Something.add_index('field1', unique=True)
    await Something.create_indexes()

    something1 = Something(field1='A', field2=':)', related=[])
    await something1.save()

    with pytest.raises(
        DuplicateKeyError,
        match="^E11000 Duplicate Key Error, full error: {'keyValue': {'field1': 'A'}, 'keyPattern': {'field1': 1}}$",
    ):
        duplicate = Something(field1='A', field2=':)', related=[])
        await duplicate.save()

    something2 = Something(field1='B', field2=';)', related=[something1.id])
    await something2.save()

    found = await Something.find_one({'field1': 'B'})
    assert found
    assert found.field1 == 'B'
    assert found.field2 == ';)'
    assert len(found.related) == 1


@pytest.mark.anyio
async def test_mongo_thingy_enums():
    connect(client_cls=AsyncMongoMockClient)

    class Importance(str, Enum):
        HIGH = 'high'
        LOW = 'low'

    class Something(AsyncThingy):
        pass

    something1 = Something(importance='high')
    await something1.save()

    assert await Something.count({'importance': Importance.HIGH})



from unittest.mock import patch
from bson import ObjectId
from pymongo.results import UpdateResult
import pytest
from mongomock_motor import AsyncMongoMockClient


@pytest.mark.anyio
async def sample_function(collection):
    result = await collection.update_one(
        filter={'_id': ObjectId()},
        update={'$set': {'field': 'value'}},
        upsert=True,
    )

    if result.acknowledged is False:
        raise RuntimeError()


def async_wrapper(value):
    async def wrapper(*args, **kwargs):
        return value
    return wrapper


@pytest.mark.anyio
async def test_patch():
    collection = AsyncMongoMockClient()['test']['test']

    with patch('mongomock_motor.AsyncMongoMockCollection.update_one', new=async_wrapper(UpdateResult({}, False))):
        with pytest.raises(RuntimeError):
            await sample_function(collection)


@pytest.mark.anyio
async def test_patch_object():
    collection = AsyncMongoMockClient()['test']['test']

    with patch.object(collection, 'update_one', new=async_wrapper(UpdateResult({}, False))):
        with pytest.raises(RuntimeError):
            await sample_function(collection)
